!=======================================================================
!                   _____________________________
!               \\//-----------------------------
!                )(   //=  //= //= //// //)) //
!               //\\ //_  //  //  /__/ //\\ //
!=======================================================================
!                   XEFFORT LIBRARY SAMPLES
!=======================================================================
! XGraph.f90 - Main application module for XGraph sample. 
! 
! XGraph sample evaluates a user-defined function using Fortran Parser
! expression evaluator by Stuart Midgley and draws a graph in Cartesian
! plane. It can display, print, copy, or export the graph in .bmp or
! .wmf format.
!
! XGraph sample demonstrates 
! - SDI architecture with a "View" window. View window is used only
!   to render an image. Such setup is used when the main window contains
!   other GUI elements, such as status bar, toolbar, or a "docked"
!   child dialog (as in this case)
! - Usage of XFTGDI routines to draw on view window's DC using
!   logical (scaled) coordinate system
! - Usage of XFTMenu routines to manage the menu
! - Note how accelerator keys (Ctrl+C, Ctrl+Shift+C are automatically handled)
! - Copying, printing, and exporting bitmaps and metafiles
!
! 2005. Xeffort.com
!     Jugoslav Dujic (jdujic@uns.ns.ac.yu)
! 
! You are free to use the code for both commercial and non-commercial
! use. The code is provided as-is without warranties. You are free to
! distribute and modify this source code provided that the list of 
! original author(s) remains untouched and your revisions are indicated
! as such. Contributions are welcome.
!=======================================================================
MODULE XGraph

USE Xeffort
USE DFWIN

IMPLICIT NONE

TYPE(X_MENU)::                xMenu                    !Main menu

!========================================================================
CONTAINS
!========================================================================
!PURPOSE: Initialization function called by XFT library on app start.
LOGICAL FUNCTION XInit(szCmdLine,nCmdShow)
!DEC$ATTRIBUTES DECORATE, ALIAS: "XINIT":: Xinit

USE View
USE FuncDlg

IMPLICIT NONE

CHARACTER(*), INTENT(IN)::    szCmdLine
INTEGER, INTENT(IN)::         nCmdShow

LOGICAL::            bSt
INTEGER::            iSt, iX, iY

INCLUDE 'Resource.fd'

!Loading menu resource
xMenu = XLoadMenu(IDR_MENU_MAIN)

bSt = XLoadAccelerators(IDR_ACCELTABLE)
!Creation of App & frame window
bSt = XCreateSDIApp(xMenu, "XGraph", IDI_ICON_MAIN, iExStyle=WS_EX_CONTROLPARENT+WS_EX_APPWINDOW)

!Handle sizing of the main window
bSt = XSetHandler(XW_FRAME, WM_SIZE, XW_FRAME_OnSize)
bSt = XSetHandler(XW_FRAME, WM_CLOSE, XW_FRAME_OnClose)

!Menu item handlers
bSt = XSetCommand(XW_FRAME, IDM_HELP_ABOUT, XFrame_OnAbout)
bSt = XSetCommand(XW_FRAME, IDM_FILE_CLOSE, XFrame_OnExit)
bSt = XSetCommand(XW_FRAME, IDM_VIEW_GRID, XW_FRAME_OnViewGrid)
bSt = XSetCommand(XW_FRAME, IDM_COPY_WMF, XW_FRAME_OnCopyWmf)
bSt = XSetCommand(XW_FRAME, IDM_COPY_BMP, XW_FRAME_OnCopyBmp)
bSt = XSetCommand(XW_FRAME, IDM_FILE_SAVEAS, XW_FRAME_OnFileSaveas)

!Create dialog for entering function properties
CALL XFuncDlg_Create()
!Create view window for image rendering
CALL XView_Create()

!Force refresh of the window/scrollbar
bSt = XGetClientRect(XW_FRAME, iX, iY)
bSt = XW_FRAME_OnSize(XW_FRAME, iX, iY, 0)

XInit = .TRUE.

END FUNCTION XInit
!========================================================================
!Handler for File/Exit menu
SUBROUTINE XFrame_OnExit(xWnd,ID,nCode)

USE FuncDlg

TYPE(X_WINDOW)::  xWnd
INTEGER::   ID, nCode

INTEGER::   iSt

iSt = XSendMessage(XW_FRAME, WM_CLOSE, 0, 0)

END SUBROUTINE XFrame_OnExit
!========================================================================
!Handler for Close button (WM_CLOSE message)
LOGICAL FUNCTION XW_FRAME_OnClose(xWnd)

USE FuncDlg

TYPE(X_WINDOW)::  xWnd
INTEGER::   ID, nCode

INTEGER::   iSt

CALL XFuncDlg_SaveRegistry()
XW_FRAME_OnClose = .FALSE.

END FUNCTION XW_FRAME_OnClose
!========================================================================
!Help/About handler
SUBROUTINE XFrame_OnAbout(xWnd,ID,nCode)

TYPE(X_WINDOW)::  xWnd
INTEGER::   ID, nCode

TYPE(X_WINDOW)::  xDlg
LOGICAL::         bSt
INTEGER::         iSt

INCLUDE "Resource.fd"

bSt = XLoadDialog(IDD_DIALOG_ABOUT, xDlg, xWnd)
iSt = XModalDialog(xDlg)
CALL XDestroyDialog(xDlg)

END SUBROUTINE XFrame_OnAbout
!=======================================================================
!Called whenever the window is resized. It demonstrates two types
!of behaviour: 
! - when the size is "locked", the virtual image size remains unchanged.
!   scrollbars appear whenever necessary
! - otherwise, the image is scaled to match the client area dimensions.
LOGICAL FUNCTION XW_FRAME_OnSize(xWnd, iWidth, iHeight, nFlag)
 
USE View
USE FuncDlg

TYPE(X_WINDOW)::              xWnd
INTEGER, INTENT(IN)::         iWidth, iHeight
INTEGER, INTENT(IN)::         nFlag

LOGICAL::                     bSt
INTEGER::                     iX, iY, jWidth, jHeight

!Move the FuncDlg window
bSt = XGetWindowPos(xFuncDlg, iX, iY, iWidth=jWidth, iHeight=jHeight)
bSt = XSetWindowPos(xFuncDlg, 0, 0, iWidth, jHeight)
!Move the View window just below the FuncDlg window
bSt = XSetWindowPos(xView, 0, jHeight, iWidth, iHeight-jHeight)
CALL XUpdateWindow(xView)

XW_FRAME_OnSize = .FALSE.

END FUNCTION XW_FRAME_OnSize
!=======================================================================
!Toggles the grid on and off and checks the menu to reflect the current
!state
SUBROUTINE XW_FRAME_OnViewGrid(xWnd, idCmd, nCode)

USE View
 
TYPE(X_WINDOW)::        xWnd
INTEGER, INTENT(IN)::   idCmd, nCode

LOGICAL::               bSt

INCLUDE "Resource.fd"

xGr%bGrid = .NOT.xGr%bGrid
bSt = XMenuSet(xMenu, idCmd, xGr%bGrid)
CALL XUpdateWindow(xView)

END SUBROUTINE XW_FRAME_OnViewGrid
!=======================================================================
!Copies the image to clipboard in metafile format
SUBROUTINE XW_FRAME_OnCopyWmf(xWnd, idCmd, nCode)

USE View
 
TYPE(X_WINDOW)::        xWnd
INTEGER, INTENT(IN)::   idCmd, nCode

TYPE(X_DC)::            xDC
TYPE(T_METAFILEPICT)::  MFP; POINTER(pMFP, MFP)
TYPE(T_POINT)::         PT
TYPE(T_SIZE)::          SZ
INTEGER::               hMem, hData, iErr, hMF
LOGICAL::               bSt

xDC%hDC = CreateMetaFile(0)

!Windows GDI MapMode must be set to MM_ANISOTROPIC to obtain scalable
!metafile. For convenience, we'll set Window's viewport and window to
!same arbitrary numbers (0,0,200,200).
bSt = SetMapMode(xDC%hDC, MM_ANISOTROPIC)
bSt = SetViewportOrgEx(xDC%hDC, 0, 0, PT)
bSt = SetViewportExtEx(xDC%hDC, 200, 200, SZ)
bSt = SetWindowOrgEx(xDC%hDC, 0, 0, PT)
bSt = SetWindowExtEx(xDC%hDC, 200, 200, SZ)

!Now, XFTGDI's viewport matches Windows GDI's viewport
CALL XSetViewport(xDC, 0, 0, 200, 200)

CALL xView_DrawGraph(xDC)

hMF = CloseMetaFile(xDC%hDC)

IF (OpenClipboard(NULL)) THEN
   hMem = GlobalAlloc(GMEM_MOVEABLE+GMEM_DDESHARE, SIZEOF(MFP))
   pMFP = GlobalLock(hMem)
   MFP%mm = MM_ISOTROPIC
   MFP%hMF = hMF

   hData = SetClipboardData(CF_METAFILEPICT, hMem)
   iErr = GetLastError()

   bSt = CloseClipboard()
END IF

END SUBROUTINE XW_FRAME_OnCopyWmf
!=======================================================================
!Copies the image to clipboard in bitmap format
SUBROUTINE XW_FRAME_OnCopyBmp(xWnd, idCmd, nCode)

USE VIEW
 
TYPE(X_WINDOW)::  xWnd
INTEGER, INTENT(IN):: idCmd, nCode

TYPE(X_DC)::            xDC
LOGICAL::               bSt
INTEGER::               iWidth, iHeight, iErr, hData, hBmp

IF (xGr%bLocked) THEN
   iWidth = xGr%iWidth
   iHeight = xGr%iHeight
ELSE
   bSt = XGetClientRect(xView, iWidth, iHeight)
END IF
xDC = XMemoryDC(iWidth, iHeight)
bSt = XFillViewport(xDC, X_BRUSH(BS_SOLID, XCOLOR_WHITE, 0))
CALL XSetViewport(xDC, 50, 10, iWidth-10, iHeight-30)

CALL XView_DrawGraph(xDC)

IF (OpenClipboard(NULL)) THEN
   hBmp = GetCurrentObject(xDC%hDC, OBJ_BITMAP)
   hData = SetClipboardData(CF_BITMAP, hBmp)
   iErr = GetLastError()
   bSt = CloseClipboard()
END IF
bSt = XDeleteDC(xDC)

END SUBROUTINE XW_FRAME_OnCopyBmp
!=======================================================================
SUBROUTINE XW_FRAME_OnFileSaveas(xWnd, idCmd, nCode)

USE View
 
TYPE(X_WINDOW)::        xWnd
INTEGER, INTENT(IN)::   idCmd, nCode

CHARACTER(MAX_PATH)::   sDir
CHARACTER(64)::         sFile
INTEGER::               iWidth, iHeight, hMf
LOGICAL::               bSt
TYPE(X_DC)::            xDC
TYPE(X_BITMAP)::        xBmp
TYPE(T_POINT)::         PT
TYPE(T_SIZE)::          SZ

IF (XGetSaveFile(xWnd, sDir, sFile, sExts=(/"*.bmp", "*.wmf"/), &
   sTypes=(/"Bitmap (.bmp)", "Windows Metafile (.wmf)"/))) THEN

   iWidth = LEN_TRIM(sFile)
   IF (TOLOWER(sFile(iWidth-2:iWidth)).EQ."bmp") THEN
      IF (xGr%bLocked) THEN
         iWidth = xGr%iWidth
         iHeight = xGr%iHeight
      ELSE
         bSt = XGetClientRect(xView, iWidth, iHeight)
      END IF
      xDC = XMemoryDC(iWidth, iHeight)
      bSt = XFillViewport(xDC, X_BRUSH(BS_SOLID, XCOLOR_WHITE, 0))
      CALL XSetViewport(xDC, 50, 10, iWidth-10, iHeight-30)

      CALL XView_DrawGraph(xDC)

      bSt = XGetBitmap(xDC, xBmp, -50, -10, iWidth-50, iHeight-10)
      bSt = XSaveBitmap(xBmp, TRIM(sDir)//"\"//sFile)

      bSt = XDeleteBitmap(xBmp)
      bSt = XDeleteDC(xDC)
   ELSE
      xDC%hDC = CreateMetaFile(TRIM(sDir)//"\"//TRIM(sFile)//CHAR(0))
      IF (xDC%hDC.NE.0) THEN
         iWidth = 400
         iHeight = 200
         bSt = SetMapMode(xDC%hDC, MM_ANISOTROPIC)
         bSt = SetViewportOrgEx(xDC%hDC, 0, 0, PT)
         bSt = SetViewportExtEx(xDC%hDC, iWidth, iHeight, SZ)
         bSt = SetWindowOrgEx(xDC%hDC, 0, 0, PT)
         bSt = SetWindowExtEx(xDC%hDC, iWidth, iHeight, SZ)
      
         CALL XSetViewport(xDC, 50, 10, iWidth-10, iHeight-30)

         CALL XView_DrawGraph(xDC)

         hMF = CloseMetaFile(xDC%hDC)
         bSt = DeleteMetaFile(hMf)
      ELSE
         hMF = GetLastError()
      END IF
   END IF
END IF


END SUBROUTINE XW_FRAME_OnFileSaveas
!========================================================================
END MODULE XGraph

