!=======================================================================
!                   _____________________________
!               \\//-----------------------------
!                )(   //=  //= //= //// //)) //
!               //\\ //_  //  //  /__/ //\\ //
!=======================================================================
!                   XEFFORT LIBRARY SAMPLES
!=======================================================================
! XView.f90 - Module for rendering of the graph specified by X_GRAPH
! structure on given window/DC
!
! 2005. Xeffort.com
!     Jugoslav Dujic (jdujic@uns.ns.ac.yu)
! 
! You are free to use the code for both commercial and non-commercial
! use. The code is provided as-is without warranties. You are free to
! distribute and modify this source code provided that the list of 
! original author(s) remains untouched and your revisions are indicated
! as such. Contributions are welcome.
!=======================================================================

MODULE View

USE XFT
USE DFWIN

IMPLICIT NONE

TYPE(X_WINDOW)::           xView

INTEGER, PARAMETER::       E_OK = 1                !Formula is valid
INTEGER, PARAMETER::       E_EMPTY = 0             !Formula is empty
INTEGER, PARAMETER::       E_ERROR = -1            !Formula contains an error

INTEGER::                  m_iStatus = E_EMPTY     !Is the formula valid?

TYPE X_GRAPH
   INTEGER::               iBkColor = XCOLOR_WHITE !Background color
   INTEGER::               iWidth, iHeight         !Graph dimensions (pixels)
   TYPE(X_WPOINT)::        wXY(500)                !Array of (x,y) points
   REAL::                  Xmin=0., Xmax=0.        !=MINVAL/MAXVAL(wXY%fX)
   REAL::                  Ymin=0., Ymax=0.        !=MINVAL/MAXVAL(wXY%fY)
   LOGICAL::               bGrid = .FALSE., &      !Draw grid
                           bLocked = .FALSE.       !Locked size
END TYPE X_GRAPH

TYPE(X_GRAPH)::            xGr                     !current graph

CHARACTER(100)::           m_sError                !Error text

!======================================================================
CONTAINS
!======================================================================
!Creates the view window.
SUBROUTINE XView_Create()

LOGICAL::      bSt
INTEGER::      iSt, iWidth, iHeight

bSt = XGetClientRect(XW_FRAME, iWidth, iHeight)
!Note XS_EX_DBLBUFFERED which eliminates flicker during resize
iSt = XCreateWindow(xView, XW_FRAME, iStyle=WS_CHILD+WS_VISIBLE+WS_CLIPSIBLINGS, &
         iExStyle=XS_EX_DBLBUFFERED,                             &
         iWidth=iWidth, iHeight=iHeight)
bSt = XSetHandler(xView, WM_PAINT, xView_OnPaint)
bSt = XSetHandler(xView, WM_VSCROLL, xView_OnScroll)
bSt = XSetHandler(xView, WM_HSCROLL, xView_OnScroll)
bSt = XSetHandler(xView, WM_WINDOWPOSCHANGING, xView_OnWindowPosChanging)

END SUBROUTINE XView_Create
!=======================================================================
!Called by Windows whenever the area has to be repainted (e.g. when
!(a portion of) the window has been revealed)
LOGICAL FUNCTION xView_OnPaint(xWnd, xDC, iX1, iY1, iX2, iY2)
 
TYPE(X_WINDOW)::              xWnd
TYPE(X_DC), INTENT(INOUT)::   xDC
INTEGER, INTENT(IN)::         iX1, iY1, iX2, iY2

LOGICAL::                     bSt
INTEGER::                     iWidth, iHeight, iPosH, iPosV

bSt = XFillViewport(xDC, X_BRUSH(BS_SOLID, xGr%iBkColor, 0))

SELECT CASE(m_iStatus)
CASE(E_OK)
   bSt = XGetClientRect(xWnd, iWidth, iHeight)
   IF (xGr%bLocked) THEN
      !If the size is locked, set the viewport with offset dependent on 
      !scrollbar position, and with extents equal to *graph* size. 
      !Additional offsets are introduced from top & left to allow for text/blank frame
      bSt=XScrollbarGet(xWnd, SB_HORZ, iPosH)
      bSt=XScrollbarGet(xWnd, SB_VERT, iPosV)
      CALL XSetViewport(xDC, 50-iPosH, 10-iPosV, xGr%iWidth-10-iPosH, xGr%iHeight-iPosV-30)
   ELSE
      !With non-locked size, set the viewport with extents equal to *window* size.
      CALL XSetViewport(xDC, 50, 10, iWidth-10, iHeight-30)
   END IF
   CALL XView_DrawGraph(xDC)
CASE(E_EMPTY)
CASE(E_ERROR)
   !Display the error text
   CALL XSetWindow(xDC, 0., 0., 100., 100.)
   bSt = XSetFont_W(xDC, "Arial", 10.)
   bSt = XTextOut_W(xDC, 50., 50., m_sError, iColor=XCOLOR_BLACK, iStyle=TA_CENTER)
END SELECT

xView_OnPaint = .TRUE.

END FUNCTION xView_OnPaint
!=======================================================================
!Renders the graph on given surface (DC) xDC.
SUBROUTINE xView_DrawGraph(xDC)

TYPE(X_DC), INTENT(INOUT)::   xDC

LOGICAL::                     bSt
INTEGER::                     i, ix1, iy1, ix2, iy2, nDecimals, nDigits
REAL::                        xMin, xMax, yMin, yMax
REAL::                        Range, RelRange, Step, XLog, TickMin, Tick

!Set the logical coordinate system to match curve's extents
CALL XSetWindow(xDC, xGr%xMin, xGr%yMax, xGr%xMax, xGr%yMin)

CALL XSetPen(xDC, XCOLOR_BLACK)
CALL XSetBrush(xDC, XCOLOR_WHITE)

bSt = XRectangle_W(xDC, xGr%xMin, xGr%yMin, xGr%xMax, xGr%yMax)

!Set the font to be approx. of 5% of graph size
CALL XGetViewport(xDC, ix1, iy1, ix2, iy2)
bSt = XSetFont(xDC, "Arial", (iy2-iy1)/20)

!This is tricky -- analyses the range and adjusts the
!tick marks accordingly. It could use some improvement, but it's
!OK for a sample project :-)
DO i=1,2
   IF (i.eq.1) THEN
      Range = xGr%xMax - xGr%xMin
   ELSE
      Range = xGr%yMax - xGr%yMin
   END IF

   XLog = LOG10(Range)
   !Scale the range into "integer" area of approx 0:10
   RelRange = Range/10.**FLOOR(XLog)
   SELECT CASE(NINT(RelRange))
   CASE(0:1)
      Step = 0.1
   CASE(2:3)
      Step = 0.2
   CASE(4:7)
      Step = 0.5
   CASE(8:)
      Step = 1
   END SELECT

   !Determine number of decimal points
   IF (XLog < 1) THEN
      nDecimals = NINT(ABS(XLog))+1
   ELSE
      nDecimals = 0
   END IF
   !and total number of digits
   nDigits=NINT(ABS(XLog))+3

   !Rescale the step back into original range
   Step = Step*(10.**FLOOR(XLog))
   !Draw the tickmarks
   CALL XSetPen(xDC, XCOLOR_LTGRAY)

   IF (i.eq.1) THEN
      !Horizontal pass
      Tick = (FLOOR(xGr%xMin/Step)+1) * Step
      DO WHILE (Tick.LE.xGr%xMax)
         IF (xGr%bGrid) THEN
            !Draw the grid line
            bSt = XLine_W(xDC, Tick, xGr%yMin, Tick, xGr%yMax)
         ELSE
            !Draw just the tickmark
            CALL XWnd2View(xDC, Tick, xGr%yMin, ix1, iy1)
            bSt = XLine(xDC, ix1, iy1, ix1, iy1+3)
         END IF
         !Tick title
         IF (SIGN(1.,Tick).LT.0) nDigits = nDigits+1
         bSt = XTextOut_W(xDC, Tick, xGr%Ymin, XString(Tick,nDecimals), iColor=XCOLOR_BLACK, iStyle=TA_CENTER)
         Tick = Tick+Step
      END DO
   ELSE
      !Vertical pass
      Tick = (FLOOR(xGr%yMin/Step)+1) * Step
      DO WHILE (Tick.LE.xGr%yMax)
         IF (xGr%bGrid) THEN
            bSt = XLine_W(xDC, xGr%xMin, Tick, xGr%xMax, Tick)
         ELSE
            !Draw the grid line
            CALL XWnd2View(xDC, xGr%xMin, Tick, ix1, iy1)
            !Draw just the tickmark
            bSt = XLine(xDC, ix1, iy1, ix1+3, iy1)
         END IF
         !Tick title
         IF (SIGN(1.,Tick).LT.0) nDigits = nDigits+1
         bSt = XTextOut_W(xDC, xGr%Xmin, Tick, XString(Tick,nDecimals), iColor=XCOLOR_BLACK, iStyle=TA_RIGHT)
         Tick = Tick+Step
      END DO
   END IF
END DO

CALL XSetPen(xDC, XCOLOR_BLACK, iWidth=2)

!Y-axis
IF (xGr%xMin<0. .and. 0.<xGr%xMax) THEN
   bSt = XLine_W(xDC, 0., xGr%yMin, 0., xGr%yMax)
END IF

!X-axis
IF (xGr%yMin<0. .and. 0.<xGr%yMax) THEN
   bSt = XLine_W(xDC, xGr%xMin, 0., xGr%xMax, 0.)
END IF

CALL XSetPen(xDC, XCOLOR_BLACK)

!And, finally, the curve itself.
bSt = XPolyLine_W(xDC, xGr%wXY, 500)

END SUBROUTINE xView_DrawGraph
!=======================================================================
!Called whenever user moves the scrollbar thumb. Note that scrollbar
!does not "stick" but it has to be set programmatically using
!XScrollbarSet
RECURSIVE LOGICAL FUNCTION xView_OnScroll(xWnd, nFlag, nScrollCode, iPos)
 
TYPE(X_WINDOW)::              xWnd
INTEGER, INTENT(IN)::         nFlag
INTEGER, INTENT(IN)::         nScrollCode
INTEGER, INTENT(IN)::         iPos

LOGICAL::                     bSt
INTEGER::                     newPos, iPage

bSt = XScrollbarGet(xWnd, nFlag, iPage=iPage)
newPos = iPos
SELECT CASE (nScrollCode)
CASE(SB_LINEUP)
   newPos = iPos - 10    !Subtract 10 pixels
CASE(SB_LINEDOWN)
   newPos = iPos + 10    !Add 10 pixels
CASE(SB_PAGEUP)
   newPos = iPos - iPage !Subtract one page
CASE(SB_PAGEDOWN)
   newPos = iPos + iPage !Add one page
CASE(SB_THUMBTRACK, SB_THUMBPOSITION)
   newPos = iPos         !Enable redrawing while the thumb is dragged
END SELECT
bSt = XScrollbarSet(xWnd, nFlag, newPos)
CALL XUpdateWindow(xWnd)

xView_OnScroll = .FALSE.

END FUNCTION xView_OnScroll
!=======================================================================
!Called whenever the view window position is changed by programmatic
!means (XMoveWindow, XSizeWindow, XSetWindowPos)
RECURSIVE LOGICAL FUNCTION xView_OnWindowPosChanging(xWnd, iX, iY, iWidth, iHeight, nFlags)
 
TYPE(X_WINDOW)::              xWnd
INTEGER, INTENT(IN)::         iX, iY
INTEGER, INTENT(IN)::         iWidth, iHeight
INTEGER, INTENT(IN)::         nFlags

LOGICAL::                     bSt
INTEGER::                     ia, ib, ic

!Set the scrollbar properties if the image is locked.
IF (IAND(nFlags,SWP_NOMOVE).EQ.0 .AND. xGr%bLocked) THEN
   bSt = XScrollbarSet(xWnd, SB_VERT, iMin=0, iMax=xGr%iHeight, iPage=iHeight)
   bSt = XScrollbarSet(xWnd, SB_HORZ, iMin=0, iMax=xGr%iWidth, iPage=iWidth)
END IF

xView_OnWindowPosChanging = .FALSE.

END FUNCTION xView_OnWindowPosChanging
!=======================================================================
END MODULE View
